import torch
import torch.nn as nn
from torch.autograd import Variable
from ..config import nn_args, shared, training

class BiGRU(nn.Module):
    """
    A class representing a Bidirectional Gated Recurrent Unit (BiGRU) model for stock prediction.

    The BiGRU model is a type of recurrent neural network (RNN) that is capable of learning long-term dependencies
    in time series data, such as stock prices. The model consists of a bidirectional GRU layer followed by a series of
    fully connected layers and activation functions. Bidirectional GRUs process the input sequence both forward and
    backward, allowing the model to capture information from both the past and the future.

    :param input_size: The number of input features for the BiGRU model.
    :type input_size: int
    :param hidden_size: The number of hidden units in each GRU layer.
    :type hidden_size: int
    :param num_layers: The number of GRU layers in the model.
    :type num_layers: int
    :param output_size: The number of output units for the BiGRU model, corresponding to the predicted target variable(s).
    :type output_size: int
    :param dropout: The dropout percentage applied between layers for regularization, preventing overfitting.
    :type dropout: float
    :example:
        >>> from stockpy.neural_network import BiGRU
        >>> bigru = BiGRU()
    """
    def __init__(self):
        super().__init__()
        self.gru = nn.GRU(input_size=nn_args.input_size, 
                          hidden_size=nn_args.hidden_size, 
                          num_layers=nn_args.num_layers, 
                          batch_first=True,
                          bidirectional=True
                          )
        self.layers = nn.Sequential(
            nn.ReLU(),
            nn.Linear(nn_args.hidden_size * 2, nn_args.hidden_size), # [2 * hidden_size] -> [hidden_size]
            nn.ReLU(),
            nn.Dropout(shared.dropout),
            nn.Linear(nn_args.hidden_size, nn_args.input_size), # [hidden_size] -> [input_size]
            nn.ReLU(),
            nn.Dropout(shared.dropout),
            nn.Linear(nn_args.input_size, nn_args.output_size), # [input_size] -> [output_size]
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Defines the forward pass of the BiGRU model.

        :param x: The input tensor.
        :type x: torch.Tensor

        :returns: The output tensor, corresponding to the predicted target variable(s).
        :rtype: torch.Tensor
        """
        batch_size = x.size(0)
        h0 = Variable(torch.zeros(nn_args.num_layers * 2, 
                                  batch_size, 
                                  nn_args.hidden_size)).to(training.device)
        
        out, _ = self.gru(x, (h0))
        out = self.layers(out[:, -1, :])       
        out = out.view(-1, 1)

        return out
    
    def to(self, device: torch.device) -> None:
        """
        Moves the model to the specified device.

        :param device: The device to move the model to.
        :type device: torch.device
        """
        super().to(device)

    @property
    def model_type(self) -> str:
        """
        Returns the type of model.

        :returns: The model type as a string.
        :rtype: str
        """
        return "neural_network"



from ..base import Base
from ._gru import GRU as _GRU
from ._bigru import BiGRU as _BiGRU
from ._bilstm import BiLSTM as _BiLSTM
from ._lstm import LSTM as _LSTM
from ._mlp import MLP as _MLP
from ..config import shared

class GRU(Base):

    def __init__(self,
                 **kwargs
                ):
        
        for key, value in kwargs.items():
            setattr(shared, key, value)

        super().__init__(model=_GRU(), **kwargs)
        
class BiGRU(Base):

    def __init__(self,
                 **kwargs
                ):
        
        for key, value in kwargs.items():
            setattr(shared, key, value)

        super().__init__(model=_BiGRU(), **kwargs)
        
class BiLSTM(Base):

    def __init__(self,
                **kwargs
                ):
    
        for key, value in kwargs.items():
            setattr(shared, key, value)

        super().__init__(model=_BiLSTM(), **kwargs)
        
class LSTM(Base):

    def __init__(self,
                **kwargs
                ):
    
        for key, value in kwargs.items():
            setattr(shared, key, value)

        super().__init__(model=_LSTM(), **kwargs)
        
class MLP(Base):

    def __init__(self,
                **kwargs
                ):
    
        for key, value in kwargs.items():
            setattr(shared, key, value)

        super().__init__(model=_MLP(), **kwargs)



import datetime
import hashlib
import os
import shutil
import sys
import glob
import sys

import torch
import torch.nn as nn
from typing import Union, Tuple

import pyro
import pyro.distributions as dist
from pyro.infer.autoguide import AutoDiagonalNormal
from pyro.infer import (
    SVI,
    Trace_ELBO,
    TraceMeanField_ELBO,
    Predictive
)
from pyro.optim import ClippedAdam
from pyro.nn import PyroModule
import torch.nn.functional as F
from ..config import nn_args, prob_args, shared, training

class Model():
    def __init__(self,
                 model=None,
                 **kwargs
                 ):
        
        if model.model_type == "neural_network":
            for key, value in kwargs.items():
                setattr(nn_args, key, value)
        elif model.model_type == "probabilistic":
            for key, value in kwargs.items():
                setattr(prob_args, key, value)        

        self._initModel(model)

    def _modelEval(self):
        print(self._model.eval())
    
    def _initModel(self, 
                   model : Union[nn.Module, PyroModule]
                   ) -> None:
        """
        Initializes the neural network model.
        Returns:
            None
        """
        self.name = model.__class__.__name__
        
        if shared.pretrained:
            path = self._initModelPath(model, self.name)
            model_dict = torch.load(path)
            model.load_state_dict(model_dict['model_state'])

        self._model = model
        self._model.to(training.device)

        if training.use_cuda:
            if torch.cuda.device_count() > 1:
                self._model = nn.DataParallel(self._model)
            
        if self._model.model_type == "neural_network":
            self.type = 'neural_network'
        elif self._model.model_type == "probabilistic":
            self.type = 'probabilistic'
            if self.name == 'BayesianNN':
                self._guide = AutoDiagonalNormal(self._model)
            else:
                self._guide = self._model.guide
                    
    def _saveModel(self, 
                   type_str : str
                   ) -> None:
        """
        Saves the model to disk.
        Parameters:
            type_str (str): a string indicating the type of model
            epoch_ndx (int): the epoch index
        Returns:
            None
        """

        file_path = os.path.join(
            '..', 
            '..', 
            'models', 
            self.name, 
            type_str + '_{}_{}.state'.format(shared.dropout,
                                                shared.weight_decay
                                                ),
            )

        os.makedirs(os.path.dirname(file_path), mode=0o755, exist_ok=True)

        model = self._model
        if isinstance(model, torch.nn.DataParallel):
            model = model.module

    def _initModelPath(self, 
                       model : Union[nn.Module, PyroModule], 
                       type_str : str) -> str:
        """
        Initializes the model path.

        Parameters:
            type_str (str): a string indicating the type of model

        Returns:
            str: the path to the initialized model
        """

        model_dir = '../../models/' + self.name
        if not os.path.exists(model_dir):
            os.makedirs(model_dir)

        local_path = os.path.join(
            '..', 
            '..', 
            'models', 
            self.name, 
            type_str + '_{}_{}.state'.format(shared.dropout,
                                                shared.weight_decay
                                                ),
            )

        file_list = glob.glob(local_path)
        
        if not file_list:
            raise ValueError(f"No matching model found in {local_path} for the given parameters.")
        
        # Return the most recent matching file
        return file_list[0]


import torch
import torch.nn as nn
from typing import Union, Tuple
import pandas as pd
import numpy as np

from .config import training
from .utils._dataloader import normalize
from .utils._training import Trainer
from .utils._model import Model
from .utils._predict import Predict
from .utils._dataloader import _initTrainValDl

class Base(Trainer, Predict):

    def __init__(self, 
                 model = None,
                 **kwargs
                 ):
    
        super().__init__(model=model, **kwargs)

    def _modelEval(self):
        print(self._model.eval())
            
    def fit(self, 
            x_train: Union[np.ndarray, pd.core.frame.DataFrame],
            **kwargs
            ) -> None:
        """
        Fits the neural network model to a given dataset.

        This method takes the input training dataset and trains the model for a specified number of epochs. 
        It uses the given batch size and sequence length to preprocess the data for the model. 
        Validation is performed at specified intervals (validation_cadence) 
        to monitor the model's performance on the validation set. 
        Early stopping is implemented using the patience parameter.

        :param x_train: The training dataset, either as a numpy array or pandas DataFrame
        :type x_train: Union[np.ndarray, pd.core.frame.DataFrame]
        :param epochs: The number of epochs to train the model for, defaults to 10
        :type epochs: int, optional
        :param sequence_length: The length of the input sequence, defaults to 30
        :type sequence_length: int, optional
        :param batch_size: The batch size to use during training, defaults to 8
        :type batch_size: int, optional
        :param num_workers: The number of workers to use for data loading, defaults to 4
        :type num_workers: int, optional
        :param validation_sequence: The number of time steps to reserve for validation during training, defaults to 30
        :type validation_sequence: int, optional
        :param validation_cadence: How often to run validation during training, defaults to 5
        :type validation_cadence: int, optional
        :param patience: How many epochs to wait for improvement in validation loss before stopping early, defaults to 5
        :type patience: int, optional

        :return: None
        """
        for key, value in kwargs.items():
            setattr(training, key, value)

        if self.name == "MLP" or self.name == "BayesianNN":
            training.sequence_length = 0
        
        train_dl, val_dl = _initTrainValDl(x_train)

        self._train(train_dl, val_dl)
                        
    def predict(self, 
                x_test: Union[np.ndarray, pd.core.frame.DataFrame]
                ) -> np.ndarray:
        """
        Generate predictions for the given test set using the trained model.

        This public method calls the internal `_predict()` method to generate predictions on the given test set. The test
        set can be provided as a NumPy array or a pandas DataFrame. The returned predictions are in the form of a NumPy
        array.

        Parameters:
            x_test (Union[np.ndarray, pd.core.frame.DataFrame]): The test set to make predictions on, either as a NumPy array or pandas DataFrame.

        Returns:
            np.ndarray: The predicted target values for the given test set, as a NumPy array.
        """
        
        return self._predict(x_test)



import torch
from pyro.infer import Predictive
import pyro.distributions as dist
import torch.nn.functional as F
import numpy as np
import pandas as pd
from ..config import training
from ._dataloader import normalize
from ._dataloader import _initValDl
from ._model import Model
from typing import Union, Tuple

class Predict(Model):
    def __init__(self, 
                 model = None,
                 **kwargs
                 ):
        super().__init__(model, **kwargs)

    def _predict(self, 
                x_test: Union[np.ndarray, pd.core.frame.DataFrame]
                ) -> np.ndarray:
        """
        Generate predictions for the given test set using the trained model.

        This method first normalizes the input test set using the same normalization method as the training set. It then
        initializes a validation DataLoader and generates predictions using either a probabilistic model (e.g., Bayesian
        neural network) or a neural network model (e.g., BiGRU), depending on the model type. The predicted values are
        rescaled back to the original scale and returned as a NumPy array.

        Parameters:
            x_test (Union[np.ndarray, pd.core.frame.DataFrame]): The test set to make predictions on, either as a NumPy array or pandas DataFrame.

        Returns:
            np.ndarray: The predicted target values for the given test set, as a NumPy array.
        """
        
        scaler = normalize(x_test)
        x_test = scaler.fit_transform()
        val_dl = _initValDl(x_test)

        if self.type == "probabilistic":
            output = self._predict_probabilistic(val_dl)
        elif self.type == "neural_network":
            output = self._predict_neural_network(val_dl)
        else:
            raise ValueError("Model type not recognized")
        output = output.to('cpu')
        output = output.detach().numpy() * scaler.std() + scaler.mean()
                    
        return output

    def _predict_neural_network(self, 
                                val_dl : torch.utils.data.DataLoader
                                ) -> torch.Tensor:
        """
        Predict target values for the given validation DataLoader using a neural network model.

        This method generates predictions using a neural network model (e.g., BiGRU). The method sets the model to evaluation
        mode, iterates over the input data in batches, and generates predictions for each batch. The predictions are then
        concatenated into a single output tensor.

        Parameters:
            val_dl (torch.utils.data.DataLoader): The validation DataLoader containing the input data to make predictions on.

        Returns:
            torch.Tensor: The predicted target values as a torch.Tensor.
        """
        output = torch.tensor([]).to(self._device)
        self._model.eval()

        with torch.no_grad():
            for x_batch, _ in val_dl:
                # to device
                x_batch = x_batch.to(self._device)
                y_star = self._model(x_batch)
                output = torch.cat((output, y_star), 0)
                
        output.to('cpu')
        return output
    
    def _predict_probabilistic(self,
                               val_dl : torch.utils.data.DataLoader
                               ) -> torch.Tensor:
        """
        Predict target values for the given validation DataLoader using a probabilistic model.

        This method generates predictions using a Bayesian Neural Network or a Deep Markov Model. For the Bayesian Neural
        Network, it uses the Pyro Predictive class to generate samples and compute the mean of the predicted values. For the
        Deep Markov Model, it computes the mean of the emission distribution for each time step and retrieves the mean for
        the last time step.

        Parameters:
            val_dl (torch.utils.data.DataLoader): The validation DataLoader containing the input data to make predictions on.

        Returns:
            torch.Tensor: The predicted target values as a torch.Tensor.
        """     
        if self.name == 'BayesianNN':
            output = torch.tensor([])
            for x_batch, _ in val_dl:
                predictive = Predictive(model=self._model, 
                                        guide=self._guide, 
                                        num_samples=training.batch_size,
                                        return_sites=("linear.weight", 
                                                        "obs", 
                                                        "_RETURN")
                                                    )
                samples = predictive(x_batch)
                site_stats = {}
                for k, v in samples.items():
                    site_stats[k] = {
                        "mean": torch.mean(v, 0)
                    }

                y_pred = site_stats['_RETURN']['mean']
                output = torch.cat((output, y_pred), 0)
            
            return output
        
        else: 
            # create a list to hold the predicted y values
            output = []

            # iterate over the test data in batches
            for x_batch, _ in val_dl:
                # make predictions for the current batch
                with torch.no_grad():
                    # compute the mean of the emission distribution for each time step
                    *_, z_loc, z_scale = self._guide(x_batch)
                    z_scale = F.softplus(z_scale)
                    z_t = dist.Normal(z_loc, z_scale).rsample()
                    mean_t, _ = self._model.emitter(z_t, x_batch)
                    
                    # get the mean for the last time step
                    mean_last = mean_t[:, -1, :]

                # add the predicted y values for the current batch to the list
                output.append(mean_last)

            # concatenate the predicted y values for all batches into a single tensor
            output = torch.cat(output)

            # reshape the tensor to get an array of shape [151,1]
            output = output.reshape(-1, 1)

            # return the predicted y values as a numpy array
            return output


import torch
import torch.nn as nn

import pyro
import pyro.distributions as dist
from pyro.nn import PyroModule
import pyro.poutine as poutine
import math
import torch.nn.functional as F
from typing import Tuple, Optional
from ..config import prob_args, training

class Emitter(nn.Module):
    """
    Parameterizes the Gaussian observation likelihood p(y_t | z_t, x_t).

    :ivar lin_z_to_hidden: a linear transformation from the latent space to a hidden state
    :vartype lin_z_to_hidden: torch.nn.Linear
    :ivar lin_x_to_hidden: a linear transformation from the input space to a hidden state
    :vartype lin_x_to_hidden: torch.nn.Linear
    :ivar lin_hidden_to_mean: a linear transformation from the hidden state to the output mean
    :vartype lin_hidden_to_mean: torch.nn.Linear
    :ivar variance: the fixed variance hyperparameter
    :vartype variance: torch.nn.Parameter
    :ivar relu: a ReLU activation function
    :vartype relu: torch.nn.ReLU

    :example:
        >>> emitter = Emitter()
        >>> print(emitter)
        Emitter(
          (lin_z_to_hidden): Linear(in_features=16, out_features=32, bias=True)
          (lin_x_to_hidden): Linear(in_features=4, out_features=32, bias=True)
          (lin_hidden_to_mean): Linear(in_features=32, out_features=1, bias=True)
          (relu): ReLU()
        )
    """

    def __init__(self):
        super().__init__()
        # initialize the three linear transformations used in the neural network
        self.lin_z_to_hidden = nn.Linear(prob_args.z_dim, prob_args.emission_dim)
        self.lin_x_to_hidden = nn.Linear(prob_args.input_size, prob_args.emission_dim)
        self.lin_hidden_to_mean = nn.Linear(prob_args.emission_dim, 1)
        # initialize the fixed variance hyperparameter
        self.variance = nn.Parameter(torch.tensor(prob_args.variance))
        # initialize the non-linearities used in the neural network
        self.relu = nn.ReLU()

    def forward(self, 
                z_t: torch.Tensor, 
                x_t: torch.Tensor) -> Tuple[torch.Tensor, torch.nn.parameter.Parameter]:
        """
        Given the latent z at a particular time step t and the input x at time step t,
        we return the mean and variance of the Gaussian distribution p(y_t|z_t, x_t).

        :param z_t: the latent variable at time step t
        :type z_t: torch.Tensor
        :param x_t: the input variable at time step t
        :type x_t: torch.Tensor
        :return: a tuple containing the mean and variance of the Gaussian distribution
        :rtype: tuple
        """

        h1 = self.relu(self.lin_z_to_hidden(z_t))
        h2 = self.relu(self.lin_x_to_hidden(x_t))
        h3 = h1 + h2  # element-wise sum of the two hidden states
        mean = self.lin_hidden_to_mean(h3)
        return mean, self.variance
    

class GatedTransition(nn.Module):
    """
    Parameterizes the dynamics of the latent variables z_t.

    :ivar lin_z_to_hidden: a linear transformation from the latent space to a hidden state
    :vartype lin_z_to_hidden: torch.nn.Linear
    :ivar lin_x_to_hidden: a linear transformation from the input space to a hidden state
    :vartype lin_x_to_hidden: torch.nn.Linear
    :ivar lin_hidden_to_hidden1: the first gated transformation
    :vartype lin_hidden_to_hidden1: torch.nn.Linear
    :ivar lin_hidden_to_hidden2: the second gated transformation
    :vartype lin_hidden_to_hidden2: torch.nn.Linear
    :ivar relu: a ReLU activation function
    :vartype relu: torch.nn.ReLU
    :ivar sigmoid: a sigmoid activation function
    :vartype sigmoid: torch.nn.Sigmoid

    :example:
        >>> gated_transition = GatedTransition()
        >>> print(gated_transition)
        GatedTransition(
          (lin_z_to_hidden): Linear(in_features=16, out_features=32, bias=True)
          (lin_x_to_hidden): Linear(in_features=4, out_features=32, bias=True)
          (lin_hidden_to_hidden1): Linear(in_features=32, out_features=32, bias=True)
          (lin_hidden_to_hidden2): Linear(in_features=32, out_features=32, bias=True)
          (relu): ReLU()
          (sigmoid): Sigmoid()
        )
    """
    def __init__(self):
        super().__init__()
        # initialize the two linear transformations used in the neural network
        self.lin_z_to_hidden = nn.Linear(prob_args.z_dim, prob_args.transition_dim)
        self.lin_x_to_hidden = nn.Linear(prob_args.input_size, prob_args.transition_dim)
        # initialize the two gated transformations used in the neural network
        self.lin_hidden_to_hidden1 = nn.Linear(prob_args.transition_dim, prob_args.transition_dim)
        self.lin_hidden_to_hidden2 = nn.Linear(prob_args.transition_dim, prob_args.transition_dim)
        # initialize the non-linearities used in the neural network
        self.relu = nn.ReLU()
        self.sigmoid = nn.Sigmoid()

    def forward(self, 
                z_t: torch.Tensor, 
                x_t: torch.Tensor) -> Tuple[torch.Tensor, torch.nn.parameter.Parameter]:
        """
        Given the latent z at a particular time step t and the input x at time step t,
        we return the parameters for the Gaussian distribution p(z_t | z_{t-1}, x_t).

        :param z_t: the latent variable at time step t
        :type z_t: torch.Tensor
        :param x_t: the input variable at time step t
        :type x_t: torch.Tensor
        :return: a tuple containing the mean and variance of the Gaussian distribution
        :rtype: tuple
        """
        h1 = self.relu(self.lin_z_to_hidden(z_t))
        h2 = self.relu(self.lin_x_to_hidden(x_t))
        gated1 = self.sigmoid(self.lin_hidden_to_hidden1(h1))
        gated2 = self.sigmoid(self.lin_hidden_to_hidden2(h2))
        h3 = gated1 * h1 + gated2 * h2
        mu_t = h3
        return mu_t, 1.0

    
class Combiner(nn.Module):
    """
    Combines the previous hidden state z_{t-1} and the current input x_t
    to produce the hidden state h_t, which is used by the emitter and
    transition networks.

    :ivar lin_z_to_hidden: a linear transformation from the latent space to a hidden state
    :vartype lin_z_to_hidden: torch.nn.Linear
    :ivar lin_rnn_to_hidden: a linear transformation from the RNN output to a hidden state
    :vartype lin_rnn_to_hidden: torch.nn.Linear
    :ivar hidden_to_loc: a linear transformation from the hidden state to the mean of the Gaussian distribution
    :vartype hidden_to_loc: torch.nn.Linear
    :ivar hidden_to_scale: a linear transformation from the hidden state to the pre-activation scale of the Gaussian distribution
    :vartype hidden_to_scale: torch.nn.Linear
    """
    def __init__(self):
        super().__init__()
        self.lin_z_to_hidden = nn.Linear(prob_args.z_dim, prob_args.emission_dim)
        self.lin_rnn_to_hidden = nn.Linear(prob_args.rnn_dim, prob_args.emission_dim)
        self.hidden_to_loc = nn.Linear(prob_args.emission_dim, prob_args.z_dim)
        self.hidden_to_scale = nn.Linear(prob_args.emission_dim, prob_args.z_dim)
        self.relu = nn.ReLU()

    def forward(self, 
                z_prev : torch.Tensor, 
                rnn_input : torch.Tensor
                ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Combines the previous hidden state z_{t-1} and the current input x_t
        to produce the hidden state h_t, which is used by the emitter and
        transition networks.

        :param z_prev: the previous latent variable
        :type z_prev: torch.Tensor
        :param rnn_input: the current input variable
        :type rnn_input: torch.Tensor
        :return: a tuple containing the mean and variance of the Gaussian distribution
        :rtype: tuple
        """
        hidden = self.relu(self.lin_z_to_hidden(z_prev) + self.lin_rnn_to_hidden(rnn_input))
        loc = self.hidden_to_loc(hidden)
        scale = F.softplus(self.hidden_to_scale(hidden))
        return loc, scale
    
class DeepMarkovModel(nn.Module):
    """
    The DeepMarkovModel class implements a Deep Markov Model, which is a
    generative model for time series data that combines recurrent neural networks
    (RNNs) and hidden Markov models (HMMs). It consists of an RNN for encoding
    input sequences, a Combiner module to combine the RNN hidden states with the
    previous latent states, and an Emitter and GatedTransition module to model
    the emission and transition probability distributions.

    :param input_size: the number of input features
    :type input_size: int
    :param rnn_dim: The number of hidden units in the RNN used for encoding the input sequences.
    :type rnn_dim: int
    :param z_dim: The dimensionality of the latent variables z.
    :type z_dim: int
    :param emission_dim: The dimensionality of the hidden state in the Emitter network.
    :type emission_dim: int
    :param transition_dim: The dimensionality of the hidden state in the Gated Transition network.
    :type transition_dim: int
    :param variance: The initial variance value for the observation distribution.
    :type variance: float

    :ivar emitter: Emitter module for modeling the emission probability distribution
    :vartype emitter: Emitter
    :ivar transition: GatedTransition module for modeling the transition probability distribution
    :vartype transition: GatedTransition
    :ivar combiner: Combiner module for combining RNN hidden states with previous latent states
    :vartype combiner: Combiner
    :ivar rnn: GRU-based RNN for encoding input sequences
    :vartype rnn: nn.GRU
    :ivar z_0: Initial latent state parameter
    :vartype z_0: nn.Parameter
    :ivar z_q_0: Initial variational latent state parameter
    :vartype z_q_0: nn.Parameter
    :ivar h_0: Initial RNN hidden state parameter
    :vartype h_0: nn.Parameter

    :example:
        >>> from stockpy.probabilistic import DeepMarkovModel
        >>> deep_markov_model = DeepMarkovModel()
    """

    def __init__(self):
        
        super().__init__()

        # instantiate PyTorch modules used in the model and guide below
        self.emitter = Emitter()
        self.transition = GatedTransition()
        self.combiner = Combiner()

        self.rnn = nn.GRU(input_size=prob_args.input_size, 
                          hidden_size=prob_args.rnn_dim,
                          batch_first=True,
                          bidirectional=False, 
                          num_layers=2
                          )

        if training.use_cuda:
            if torch.cuda.device_count() > 1:
                self.emitter = nn.DataParallel(self.emitter)
                self.transition = nn.DataParallel(self.transition)
                self.combiner = nn.DataParallel(self.combiner)
                self.rnn = nn.DataParallel(self.rnn)

            self.emitter = self.emitter.to(training.device)
            self.transition = self.transition.to(training.device)
            self.combiner = self.combiner.to(training.device)
            self.rnn = self.rnn.to(training.device)

        # define a (trainable) parameters z_0 and z_q_0 that help define
        # the probability distributions p(z_1) and q(z_1)
        # (since for t = 1 there are no previous latents to condition on)
        self.z_0 = nn.Parameter(torch.zeros(prob_args.z_dim))
        self.z_q_0 = nn.Parameter(torch.zeros(1, prob_args.z_dim))
        # define a (trainable) parameter for the initial hidden state of the RNN
        self.h_0 = nn.Parameter(torch.zeros(1, 1, prob_args.rnn_dim))

    def model(self, 
              x_data: torch.Tensor, 
              y_data: Optional[torch.Tensor] = None, 
              annealing_factor: float = 1.0
              ) -> torch.Tensor:
        """
        This function defines the generative process of the Deep Markov Model (DMM). 
        It takes a mini-batch of input sequences `x_data` and, if available, their corresponding
        output sequences `y_data`, and generates a sequence of latent variables `z_t` at each time step.
        It uses an RNN to encode the input sequences, and a Combiner module to combine the RNN hidden states
        with the previous latent states. The Emitter and GatedTransition modules are then used to model the
        emission and transition probability distributions, respectively. It returns the final sampled latent
        variable `z_t`.
        
        :param x_data: input data of size (batch_size, num_time_steps, input_size)
        :type x_data: torch.Tensor
        :param y_data: output data of size (batch_size, num_time_steps, output_size), defaults to None
        :type y_data: Optional[torch.Tensor], optional
        :param annealing_factor: controls the weight of the KL divergence term, defaults to 1.0
        :type annealing_factor: float, optional

        :returns: tensor of size (batch_size, latent_dim) representing the final sampled latent variable
        :rtype: torch.Tensor
        
        """

        # this is the number of time steps we need to process in the data
        T_max = x_data.size(1)

        # register all PyTorch (sub)modules with pyro
        # this needs to happen in both the model and guide
        pyro.module("dmm", self)

        # set z_prev = z_0 to setup the recursive conditioning in p(z_t | z_{t-1})
        z_prev = self.z_0.expand(x_data.size(0), self.z_0.size(0))

        # we enclose all the sample statements in the model in a plate.
        # this marks that each datapoint is conditionally independent of the others
        with pyro.plate("z_minibatch", len(x_data)):
            # sample the latents z and observed y's one time step at a time
            for t in range(1, T_max + 1):
                # the next chunk of code samples z_t ~ p(z_t | z_{t-1})
                # note that (both here and elsewhere) we use poutine.scale to take care
                # of KL annealing.

                # first compute the parameters of the diagonal gaussian
                # distribution p(z_t | z_{t-1})
                z_loc, z_scale = self.transition(z_prev, x_data[:, t - 1, :])

                # then sample z_t according to dist.Normal(z_loc, z_scale).
                # note that we use the reshape method so that the univariate
                # Normal distribution is treated as a multivariate Normal
                # distribution with a diagonal covariance.
                with poutine.scale(None, annealing_factor):
                    z_t = pyro.sample("z_%d" % t,
                                    dist.Normal(z_loc, z_scale)
                                                .to_event(1))

                # compute the mean of the Gaussian distribution p(y_t | z_t)
                mean_t, variance = self.emitter(z_t, x_data[:, t - 1, :])

                # the next statement instructs pyro to observe y_t according to the
                # Gaussian distribution p(y_t | z_t)
                pyro.sample("obs_y_%d" % t,
                            dist.Normal(mean_t, variance).to_event(1),
                            obs=y_data)
                
                # print(f"model z_prev shape: {z_prev.shape}")
                # print(f"model x_data[:, t - 1, :] shape: {x_data[:, t - 1, :].shape}")   
                # the latent sampled at this time step will be conditioned upon
                # in the next time step so keep track of it
                z_prev = z_t
            
            return z_t
            
    def guide(self,
              x_data: torch.Tensor, 
              y_data: Optional[torch.Tensor] = None, 
              annealing_factor: float = 1.0
              ) -> torch.Tensor:
        """
        This function defines the variational approximation (inference network) of the Deep Markov Model (DMM). 
        It takes a mini-batch of input sequences `x_data` and, if available, their corresponding
        output sequences `y_data`, and generates a sequence of latent variables `z_t` at each time step.
        The guide function aims to approximate the true posterior distribution over the latent variables
        given the input sequences. It uses an RNN to encode the input sequences, and a Combiner module to 
        combine the RNN hidden states with the previous latent states. It returns the final sampled latent
        variable `z_t`.

        :param x_data: The input data tensor of shape (batch_size, T_max, input_size)
        :type x_data: torch.Tensor
        :param y_data: The output data tensor of shape (batch_size, T_max, output_size), defaults to None
        :type y_data: Optional[torch.Tensor], optional
        :param annealing_factor: The scaling factor for the KL term in the loss function, defaults to 1.0
        :type annealing_factor: float, optional

        :returns: The latent variable tensor sampled at the final time step, of shape (batch_size, z_dim)
        :rtype: torch.Tensor
        """
        
        # this is the number of time steps we need to process in the mini-batch
        T_max = x_data.size(1)
        # register all PyTorch (sub)modules with pyro
        pyro.module("dmm", self)
        
        # if on gpu we need the fully broadcast view of the rnn initial state
        # to be in contiguous gpu memory
        h_0_contig = self.h_0.expand(2, x_data.size(0), self.rnn.hidden_size).contiguous()
        
        # push the observed x's through the rnn;
        # rnn_output contains the hidden state at each time step
        rnn_output, _ = self.rnn(x_data, h_0_contig)
        
        z_q_0_expanded = self.z_q_0.expand(x_data.size(0), prob_args.z_dim)
        z_prev = z_q_0_expanded
        
        # we enclose all the sample statements in the guide in a plate.
        # this marks that each datapoint is conditionally independent of the others.
        with pyro.plate("z_minibatch", len(x_data)):
            # sample the latents z one time step at a time
            for t in range(1, T_max + 1):
                # the next two lines assemble the distribution q(z_t | z_{t-1}, x_{t:T})

                z_loc, z_scale = self.combiner(z_prev, rnn_output[:, t - 1, :])
                z_dist = dist.Normal(z_loc, z_scale).to_event(1)

                # sample z_t from the distribution z_dist
                with pyro.poutine.scale(None, annealing_factor):
                    z_t = pyro.sample("z_%d" % t, 
                                      z_dist)
           
                # the latent sampled at this time step will be conditioned
                # upon in the next time step so keep track of it
                z_prev = z_t

            return z_t

    @property
    def model_type(self) -> str:
        """
        Returns the type of model.

        :returns: The model type as a string.
        :rtype: str
        """
        return "probabilistic"


import torch
import torch.nn as nn
from torch.autograd import Variable
from ..config import nn_args, shared, training

class BiLSTM(nn.Module):
    """
    A class representing a Bidirectional Long Short-Term Memory (BiLSTM) model for stock prediction.

    The BiLSTM model extends the LSTM model by processing the input data in both forward and backward directions,
    allowing it to capture both past and future dependencies in time series data, such as stock prices.
    It consists of a bidirectional LSTM layer followed by a series of fully connected layers and activation functions.

    :param input_size: The number of input features for the BiLSTM model.
    :type input_size: int
    :param hidden_size: The number of hidden units in each LSTM layer.
    :type hidden_size: int
    :param num_layers: The number of LSTM layers in the model.
    :type num_layers: int
    :param output_size: The number of output units for the BiLSTM model, corresponding to the predicted target variable(s).
    :type output_size: int
    :param dropout: The dropout percentage applied between layers for regularization, preventing overfitting.
    :type dropout: float
    :example:
        >>> from stockpy.neural_network import BiLSTM
        >>> bilstm = BiLSTM()
    """
    def __init__(self):
        super().__init__()
        self.lstm = nn.LSTM(input_size=nn_args.input_size, 
                            hidden_size=nn_args.hidden_size, 
                            num_layers=nn_args.num_layers, 
                            batch_first=True,
                            bidirectional=True
                            )
        self.layers = nn.Sequential(
            nn.ReLU(),
            nn.Linear(nn_args.hidden_size * 2, nn_args.hidden_size), # [2 * hidden_size] -> [hidden_size]
            nn.ReLU(),
            nn.Dropout(shared.dropout),
            nn.Linear(nn_args.hidden_size, nn_args.input_size), # [hidden_size] -> [input_size]
            nn.ReLU(),
            nn.Dropout(shared.dropout),
            nn.Linear(nn_args.input_size, nn_args.output_size), # [input_size] -> [output_size]
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Defines the forward pass of the BiLSTM model.

        :param x: The input tensor.
        :type x: torch.Tensor

        :returns: The output tensor, corresponding to the predicted target variable(s).
        :rtype: torch.Tensor
        """
        batch_size = x.size(0)
        h0 = Variable(torch.zeros(nn_args.num_layers * 2, 
                                  batch_size, 
                                  nn_args.hidden_size)
                                  ).to(self.device)
        c0 = Variable(torch.zeros(nn_args.num_layers * 2, 
                                  batch_size, 
                                  nn_args.hidden_size)).to(training.device)
        
        out, _ = self.lstm(x, (h0, c0))
        out = self.layers(out[:, -1, :])       
        out = out.view(-1, 1)

        return out
    
    def to(self, device: torch.device) -> None:
        """
        Moves the model to the specified device.

        :param device: The device to move the model to.
        :type device: torch.device
        """
        super().to(device)
    
    @property
    def model_type(self) -> str:
        """
        Returns the type of model.

        :returns: The model type as a string.
        :rtype: str
        """
        return "neural_network"



import torch
from dataclasses import dataclass

@dataclass
class nn_args:
    # neural network
    input_size: int = 4
    hidden_size: int = 8
    output_size: int = 1
    num_layers: int = 2

@dataclass
class prob_args:
    # probabilistic
    input_size : int = 4
    hidden_size : int = 8
    output_size: int = 1
    rnn_dim: int = 32
    z_dim: int = 32
    emission_dim: int = 32
    transition_dim: int = 32
    variance: float = 0.1

@dataclass
class shared:
    # shared
    dropout: float = 0.2
    pretrained: bool = False
    lr: float = 0.001
    betas: tuple = (0.9, 0.999)
    lrd: float = 0.99996
    clip_norm: float = 10.0
    weight_decay: float = 0.001
    eps: float = 1e-8
    amsgrad: bool = False
    optim_args: float = 0.01
    gamma: float = 0.1
    step_size: float = 50

@dataclass
class training:
    # training
    epochs : int = 10
    batch_size : int = 24
    sequence_length : int = 30
    validation_sequence : int = 30
    use_cuda = torch.cuda.is_available()
    device : str = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    num_workers : int = 4
    validation_cadence : int = 5
    patience : int = 5
    



import torch
import torch.nn as nn
import math

import pyro
import pyro.distributions as dist
from pyro.nn import PyroModule
import pyro.poutine as poutine
from typing import Tuple, Optional
from ..config import prob_args, training

class Emitter(nn.Module):
    """
    Parameterizes the Gaussian observation likelihood p(y_t | z_t, x_t).

    :ivar lin_z_to_hidden: a linear transformation from the latent space to a hidden state
    :vartype lin_z_to_hidden: torch.nn.Linear
    :ivar lin_x_to_hidden: a linear transformation from the input space to a hidden state
    :vartype lin_x_to_hidden: torch.nn.Linear
    :ivar lin_hidden_to_mean: a linear transformation from the hidden state to the output mean
    :vartype lin_hidden_to_mean: torch.nn.Linear
    :ivar variance: the fixed variance hyperparameter
    :vartype variance: torch.nn.Parameter
    :ivar relu: a ReLU activation function
    :vartype relu: torch.nn.ReLU

    :example:
        >>> emitter = Emitter()
        >>> print(emitter)
        Emitter(
          (lin_z_to_hidden): Linear(in_features=16, out_features=32, bias=True)
          (lin_x_to_hidden): Linear(in_features=4, out_features=32, bias=True)
          (lin_hidden_to_mean): Linear(in_features=32, out_features=1, bias=True)
          (relu): ReLU()
        )
    """

    def __init__(self):
        super().__init__()
        # initialize the three linear transformations used in the neural network
        self.lin_z_to_hidden = nn.Linear(prob_args.z_dim, prob_args.emission_dim)
        self.lin_x_to_hidden = nn.Linear(prob_args.input_size, prob_args.emission_dim)
        self.lin_hidden_to_mean = nn.Linear(prob_args.emission_dim, 1)
        # initialize the fixed variance hyperparameter
        self.variance = nn.Parameter(torch.tensor(prob_args.variance))
        # initialize the non-linearities used in the neural network
        self.relu = nn.ReLU()

    def forward(self, 
                z_t: torch.Tensor, 
                x_t: torch.Tensor) -> Tuple[torch.Tensor, torch.nn.parameter.Parameter]:
        """
        Given the latent z at a particular time step t and the input x at time step t,
        we return the mean and variance of the Gaussian distribution p(y_t|z_t, x_t).

        :param z_t: the latent variable at time step t
        :type z_t: torch.Tensor
        :param x_t: the input variable at time step t
        :type x_t: torch.Tensor
        :return: a tuple containing the mean and variance of the Gaussian distribution
        :rtype: tuple
        """

        h1 = self.relu(self.lin_z_to_hidden(z_t))
        h2 = self.relu(self.lin_x_to_hidden(x_t))
        h3 = h1 + h2  # element-wise sum of the two hidden states
        mean = self.lin_hidden_to_mean(h3)
        return mean, self.variance
    

class GatedTransition(nn.Module):
    """
    Parameterizes the dynamics of the latent variables z_t.

    :ivar lin_z_to_hidden: a linear transformation from the latent space to a hidden state
    :vartype lin_z_to_hidden: torch.nn.Linear
    :ivar lin_x_to_hidden: a linear transformation from the input space to a hidden state
    :vartype lin_x_to_hidden: torch.nn.Linear
    :ivar lin_hidden_to_hidden1: the first gated transformation
    :vartype lin_hidden_to_hidden1: torch.nn.Linear
    :ivar lin_hidden_to_hidden2: the second gated transformation
    :vartype lin_hidden_to_hidden2: torch.nn.Linear
    :ivar relu: a ReLU activation function
    :vartype relu: torch.nn.ReLU
    :ivar sigmoid: a sigmoid activation function
    :vartype sigmoid: torch.nn.Sigmoid

    :example:
        >>> gated_transition = GatedTransition()
        >>> print(gated_transition)
        GatedTransition(
          (lin_z_to_hidden): Linear(in_features=16, out_features=32, bias=True)
          (lin_x_to_hidden): Linear(in_features=4, out_features=32, bias=True)
          (lin_hidden_to_hidden1): Linear(in_features=32, out_features=32, bias=True)
          (lin_hidden_to_hidden2): Linear(in_features=32, out_features=32, bias=True)
          (relu): ReLU()
          (sigmoid): Sigmoid()
        )
    """
    def __init__(self):
        super().__init__()
        # initialize the two linear transformations used in the neural network
        self.lin_z_to_hidden = nn.Linear(prob_args.z_dim, prob_args.transition_dim)
        self.lin_x_to_hidden = nn.Linear(prob_args.input_size, prob_args.transition_dim)
        # initialize the two gated transformations used in the neural network
        self.lin_hidden_to_hidden1 = nn.Linear(prob_args.transition_dim, prob_args.transition_dim)
        self.lin_hidden_to_hidden2 = nn.Linear(prob_args.transition_dim, prob_args.transition_dim)
        # initialize the non-linearities used in the neural network
        self.relu = nn.ReLU()
        self.sigmoid = nn.Sigmoid()

    def forward(self, 
                z_t: torch.Tensor, 
                x_t: torch.Tensor) -> Tuple[torch.Tensor, torch.nn.parameter.Parameter]:
        """
        Given the latent z at a particular time step t and the input x at time step t,
        we return the parameters for the Gaussian distribution p(z_t | z_{t-1}, x_t).

        :param z_t: the latent variable at time step t
        :type z_t: torch.Tensor
        :param x_t: the input variable at time step t
        :type x_t: torch.Tensor
        :return: a tuple containing the mean and variance of the Gaussian distribution
        :rtype: tuple
        """
        h1 = self.relu(self.lin_z_to_hidden(z_t))
        h2 = self.relu(self.lin_x_to_hidden(x_t))
        gated1 = self.sigmoid(self.lin_hidden_to_hidden1(h1))
        gated2 = self.sigmoid(self.lin_hidden_to_hidden2(h2))
        h3 = gated1 * h1 + gated2 * h2
        mu_t = h3
        return mu_t, 1.0

class GaussianHMM(nn.Module):
    """
    This PyTorch Module encapsulates the model as well as the variational distribution (the guide)
    for the Gaussian Hidden Markov Model.

    :param input_size: the number of input features
    :type input_size: int
    :param z_dim: The dimensionality of the latent variables z.
    :type z_dim: int
    :param emission_dim: The dimensionality of the hidden state in the Emitter network.
    :type emission_dim: int
    :param transition_dim: The dimensionality of the hidden state in the Gated Transition network.
    :type transition_dim: int
    :param variance: The initial variance value for the observation distribution.
    :type variance: float

    :ivar emitter: the Emitter module which parameterizes the Gaussian observation likelihood p(y_t | z_t, x_t)
    :vartype emitter: Emitter
    :ivar transition: the GatedTransition module which parameterizes the dynamics of the latent variables z_t
    :vartype transition: GatedTransition
    :ivar z_0: a trainable parameter representing the initial latent state
    :vartype z_0: torch.nn.Parameter
    :ivar z_q_0: a trainable parameter representing the initial latent state for the variational distribution
    :vartype z_q_0: torch.nn.Parameter
    :example:
        >>> from stockpy.probabilistic import GaussianHMM
        >>> gaussian_hmm = GaussianHMM()
    """

    def __init__(self):
        
        super().__init__()

        use_cuda = torch.cuda.is_available()
        device = torch.device("cuda" if use_cuda else "cpu")

        # instantiate PyTorch modules used in the model and guide below
        self.emitter = Emitter()
        self.transition = GatedTransition()

        if training.use_cuda:
            if torch.cuda.device_count() > 1:
                self.emitter = nn.DataParallel(self.emitter)
                self.transition = nn.DataParallel(self.transition)

            self.emitter = self.emitter.to(training.device)
            self.transition = self.transition.to(training.device)

        # define a (trainable) parameters z_0 and z_q_0 that help define
        # the probability distributions p(z_1) and q(z_1)
        # (since for t = 1 there are no previous latents to condition on)
        self.z_0 = nn.Parameter(torch.zeros(prob_args.z_dim))
        self.z_q_0 = nn.Parameter(torch.zeros(1, prob_args.z_dim))

    def model(self, 
              x_data: torch.Tensor, 
              y_data: Optional[torch.Tensor] = None, 
              annealing_factor: float = 1.0
              ) -> torch.Tensor:
        """
        This function defines the generative process of the Gaussian Hidden Markov Model (GaussianHMM).
        It takes a mini-batch of input sequences `x_data`, and if available, their corresponding
        output sequences `y_data`, and generates a sequence of latent variables `z_t` at each time step.
        It uses a DNN to model the emission probability distribution `p(y_t|z_t, x_t)` and 
        another DNN to model the transition probability distribution `p(z_t|z_{t-1}, x_t)`.
        It returns the final sampled latent variable `z_t`.

        :param x_data: input data of size (batch_size, num_time_steps, input_size)
        :type x_data: torch.Tensor
        :param y_data: output data of size (batch_size, num_time_steps, output_size), defaults to None
        :type y_data: Optional[torch.Tensor], optional
        :param annealing_factor: controls the weight of the KL divergence term, defaults to 1.0
        :type annealing_factor: float, optional

        :returns: tensor of size (batch_size, latent_dim) representing the final sampled latent variable
        :rtype: torch.Tensor
        """

        T_max = x_data.size(1)

        # register all PyTorch (sub)modules with pyro
        # this needs to happen in both the model and guide
        pyro.module("ghmm", self)

        # set z_prev = z_0 to setup the recursive conditioning in p(z_t | z_{t-1})
        z_prev = self.z_0.expand(x_data.size(0), self.z_0.size(0))

        # we enclose all the sample statements in the model in a plate.
        # this marks that each datapoint is conditionally independent of the others
        with pyro.plate("z_minibatch", len(x_data)):
            # sequentially sample the remaining latent and observed variables
            for t in range(1, T_max):
                # sample the next latent state z_t from the transition model
                z_loc, z_scale = self.transition(z_prev, x_data[:, t - 1, :])

                with poutine.scale(None, annealing_factor):
                    z_t = pyro.sample("z_%d" % t,
                                    dist.Normal(z_loc, z_scale)
                                                .to_event(1))

                # compute the mean of the Gaussian distribution p(y_t | z_t)
                mean_t, variance = self.emitter(z_t, x_data[:, t - 1, :])
                # the next statement instructs pyro to observe y_t according to the
                # Gaussian distribution p(y_t | z_t)
                pyro.sample("obs_y_%d" % t,
                            dist.Normal(mean_t, variance).to_event(1),
                            obs=y_data)
                
                # print(f"model z_prev shape: {z_prev.shape}")
                # print(f"model x_data[:, t - 1, :] shape: {x_data[:, t - 1, :].shape}")   
                # the latent sampled at this time step will be conditioned upon
                # in the next time step so keep track of it
                z_prev = z_t
        
        return z_t

    def guide(self, 
              x_data: torch.Tensor, 
              y_data: Optional[torch.Tensor] = None, 
              annealing_factor: float = 1.0
              ) -> torch.Tensor:
        """
        Defines the guide function for the Deep Markov Model. This function is used
        for the variational inference procedure in Pyro.

        :param x_data: The input data tensor of shape (batch_size, T_max, input_size)
        :type x_data: torch.Tensor
        :param y_data: The output data tensor of shape (batch_size, T_max, output_size), defaults to None
        :type y_data: Optional[torch.Tensor], optional
        :param annealing_factor: The scaling factor for the KL term in the loss function, defaults to 1.0
        :type annealing_factor: float, optional

        :returns: The latent variable tensor sampled at the final time step, of shape (batch_size, z_dim)
        :rtype: torch.Tensor
        """
        
        T_max = x_data.size(1)

        # register all PyTorch (sub)modules with pyro
        # this needs to happen in both the model and guide
        pyro.module("ghmm", self)

        # initialize the values of the latent variables using heuristics
        z_q_0_expanded = self.z_q_0.expand(x_data.size(0), prob_args.z_dim)
        z_prev = z_q_0_expanded

        with pyro.plate("z_minibatch", len(x_data)):
            # mark z_0 as auxiliary
            for t in range(T_max):
                # define the distribution q(z_t | z_{t-1}, x_{t:T})
                z_loc, z_scale = self.transition(z_prev, x_data[:, t - 1, :])
                z_dist = dist.Normal(z_loc, z_scale).to_event(1)

                # sample z_t from the distribution z_dist
                with pyro.poutine.scale(None, annealing_factor):
                    z_t = pyro.sample("z_%d" % t, 
                                      z_dist)

                # the latent sampled at this time step will be conditioned
                # upon in the next time step so keep track of it
                z_prev = z_t

            return z_t

    @property
    def model_type(self):
        return "probabilistic"


from ._dataloader import StockDataset
from ._dataloader import normalize
from ._dataloader import StockData
from ._dataloader import _initTrainDl
from ._dataloader import _initValDl
from ._dataloader import _initTrainValDl
from ._training import Trainer
from ._model import Model
from ._predict import Predict



import torch
import torch.nn as nn
from torch.autograd import Variable
from ..config import nn_args, shared, training

class LSTM(nn.Module):
    """
    A class representing a Long Short-Term Memory (LSTM) model for stock prediction.

    The LSTM model is designed to capture long-range dependencies in time series data, such as stock prices.
    It consists of an LSTM layer followed by a series of fully connected layers and activation functions.

    :param input_size: The number of input features for the LSTM model.
    :type input_size: int
    :param hidden_size: The number of hidden units in each LSTM layer.
    :type hidden_size: int
    :param num_layers: The number of LSTM layers in the model.
    :type num_layers: int
    :param output_size: The number of output units for the LSTM model, corresponding to the predicted target variable(s).
    :type output_size: int
    :param dropout: The dropout percentage applied between layers for regularization, preventing overfitting.
    :type dropout: float
    :example:
        >>> from stockpy.neural_network import LSTM
        >>> lstm = LSTM()
    """
    def __init__(self):
        super().__init__()
        self.lstm = nn.LSTM(input_size=nn_args.input_size, 
                            hidden_size=nn_args.hidden_size * 2, 
                            num_layers=nn_args.num_layers, 
                            batch_first=True)
        
        self.layers = nn.Sequential(
            nn.ReLU(),
            nn.Linear(nn_args.hidden_size * 2, nn_args.hidden_size), # [2 * hidden_size] -> [hidden_size]
            nn.ReLU(),
            nn.Dropout(shared.dropout),
            nn.Linear(nn_args.hidden_size, nn_args.input_size), # [hidden_size] -> [input_size]
            nn.ReLU(),
            nn.Dropout(shared.dropout),
            nn.Linear(nn_args.input_size, nn_args.output_size), # [input_size] -> [output_size]
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Defines the forward pass of the LSTM model.

        :param x: The input tensor.
        :type x: torch.Tensor

        :returns: The output tensor, corresponding to the predicted target variable(s).
        :rtype: torch.Tensor
        """
        batch_size = x.size(0)
        h0 = Variable(torch.zeros(nn_args.num_layers, batch_size, nn_args.hidden_size * 2)).to(training.device)
        c0 = Variable(torch.zeros(nn_args.num_layers, batch_size, nn_args.hidden_size * 2)).to(training.device)
        _, (hn, _) = self.lstm(x, (h0, c0))
        out = self.layers(hn[0])   
        out = out.view(-1,1)

        return out

    def to(self, device: torch.device) -> None:
        """
        Moves the model to the specified device.

        :param device: The device to move the model to.
        :type device: torch.device
        """
        super().to(device)

    
    @property
    def model_type(self) -> str:
        """
        Returns the type of model.

        :returns: The model type as a string.
        :rtype: str
        """
        return "neural_network"


import torch
import torch.nn as nn

import pyro
import pyro.distributions as dist
from pyro.nn import PyroModule
from ..config import prob_args, shared


class BayesianNN(PyroModule):
    """
    This class implements a Bayesian Neural Network model using Pyro. 
    It consists of three linear layers with ReLU activation and dropout applied between them. 
    The final layer uses a Student's t-distribution instead of a Normal distribution. 
    The model is trained using maximum likelihood estimation.

    :param input_size: the number of input features
    :type input_size: int
    :param hidden_size: the number of hidden units in the GRU layer
    :type hidden_size: int
    :param num_layers: the number of GRU layers
    :type num_layers: int
    :param output_dim: the number of output units
    :type output_dim: int

    :ivar layers: a PyroModule containing three linear layers with ReLU activation and dropout applied between them.
    :vartype layers: PyroModule[nn.Sequential]

    :example:
        >>> from stockpy.probabilistic import BayesianNN
        >>> bayesian_nn = BayesianNN()
    """
    
    def __init__(self):
        """
        Initializes the Bayesian Neural Network model.

        :param prob_args: a class containing the model hyperparameters
        :type prob_args: ModelArgs
        """
        
        super().__init__()
        self.layers = PyroModule[nn.Sequential](
            PyroModule[nn.Linear](prob_args.input_size, 
                                  prob_args.hidden_size), # [4] -> [8]
            PyroModule[nn.ReLU](),
            PyroModule[nn.Dropout](shared.dropout),
            PyroModule[nn.Linear](prob_args.hidden_size, 
                                  prob_args.input_size), # [8] -> [4]
            PyroModule[nn.ReLU](),
            PyroModule[nn.Dropout](shared.dropout),
            PyroModule[nn.Linear](prob_args.input_size, 
                                  prob_args.output_size), # [4] -> [1]
        )

    def forward(self, x_data: torch.Tensor, 
                y_data: torch.Tensor=None) -> torch.Tensor:
        """
        Computes the forward pass of the Bayesian Neural Network model using Pyro.

        :param x_data: the input data tensor
        :type x_data: torch.Tensor
        :param y_data: the target data tensor
        :type y_data: torch.Tensor

        :returns: the output tensor of the model
        :rtype: torch.Tensor
        """
        x =  self.layers(x_data)
        # use StudentT distribution instead of Normal
        df = pyro.sample("df", dist.Exponential(1.))
        scale = pyro.sample("scale", dist.HalfCauchy(2.5))
        with pyro.plate("data", x_data.shape[0]):
            obs = pyro.sample("obs", dist.StudentT(df, x, scale).to_event(1), 
                              obs=y_data)
            
        return x

    @property
    def model_type(self):
        """
        Returns the type of the model.

        :returns: the type of the model
        :rtype: str
        """
        return "probabilistic"



from ..base import Base
from ._bnn import BayesianNN as BNN
from ._ghmm import GaussianHMM as GHMM 
from ._dmm import DeepMarkovModel as DMM
from ..config import shared

class BayesianNN(Base):
    def __init__(self,
                 **kwargs
                ):
        
        for key, value in kwargs.items():
            setattr(shared, key, value)

        super().__init__(model=BNN(), **kwargs)

class GaussianHMM(Base):
    def __init__(self,
                 **kwargs
                ):
        
        for key, value in kwargs.items():
            setattr(shared, key, value)

        super().__init__(model=GHMM(), **kwargs)

class DeepMarkovModel(Base):
    def __init__(self,
                 **kwargs
                ):
        
        for key, value in kwargs.items():
            setattr(shared, key, value)

        super().__init__(model=DMM(), **kwargs)



import math
import numpy as np
import pandas as pd
import os
import glob
import yfinance as yf
from yahoofinancials import YahooFinancials
from pandas_datareader import data as pdr
import pathlib
from tqdm.auto import tqdm
from torch.utils.data import DataLoader
from datetime import date, datetime
from typing import Union, Tuple

import torch
from torch import nn as nn
from sklearn.preprocessing import StandardScaler
from dataclasses import dataclass
from ..config import training


class StockData():

    def __init__(self, 
                download=False,
                download_stock=None,
                start=None,
                end=None,
                update=False,
                update_stock=None,
                delete=False,
                delete_stock=None,
                range=0,
                folder="../../stock/"):

        self._download = download
        self._download_stock = download_stock
        self._start = start
        self._end = end
        self._update = update
        self._update_stock = update_stock
        self._delete = delete
        self._delete_stock = delete_stock
        self._range = range
        self._folder = folder

        if self._range > 0 or type(self._download_stock) is not list:
            self.stock_market = pd.read_html('https://en.wikipedia.org/wiki/List_of_S%26P_500_companies')[0]['Symbol']
            self.stock_market = self.stock_market[:self._range]

        else:
            if type(self._download_stock) is list:
                self.stock_market = [i.upper() for i in self._download_stock]
            else:
                self.stock_market = pd.read_html('https://en.wikipedia.org/wiki/List_of_S%26P_500_companies')[0]['Symbol']
        
        self.__main()

    def __download(self):

        if self._end == "today" or self._end is None:
            self._end = date.today().strftime("%Y-%m-%d")
        if self._start is None:
            self._start = "2017-01-01"

        yf.pdr_override()
        
        if type(self._download_stock) is not list:
            assert self._download_stock.find(",") == -1, "Wrong input format"
            if self._download_stock is not None:
                self._download_stock = self._download_stock.upper()
                self._download_stock = self._download_stock.replace('=', '')
                self._download_stock = self._download_stock.replace(' ', '')
                if self._download_stock.find('.') != -1:
                    self._download_stock = self._download_stock.replace('.', '-') 
                    
                stock_data = pdr.get_data_yahoo(self._download_stock, 
                                                start=self._start, 
                                                end=self._end,
                                                progress=True,
                                                threads=False,
                                                ignore_tz = False
                                                )
                stock_data.to_csv(self._folder + self._download_stock + ".csv", \
                                    index=True)

        else: 
            with tqdm(total=len(self.stock_market), leave=False) as bar:
                for i in self.stock_market:
                    bar.update()
                    if i.find('.') != -1:
                        i = i.replace('.', '-')
                    stock_data = pdr.get_data_yahoo(i, 
                                            start=self._start, 
                                            end=self._end,
                                            progress=False,
                                            threads=True,
                                            ignore_tz = False,
                                            )

                    stock_data.to_csv(self._folder + i + ".csv", 
                                                index=True)

    def download_stock(self):

        if type(self._download_stock) is list: 
            for i in self._download_stock:
                path = self._folder + i.upper() + '.csv'

                if os.path.isfile(path) is False or len(os.listdir(self._folder)) == 0:
                    return self.__download()
                else:
                    check = input('This file already exist, do you want download proceed [y/n]: ')
                    if check.lower() == 'y':
                        return self.__download()
                    else: raise Exception("download failed")
        else:        
            self._download_stock = self._download_stock.upper() 
            path = self._folder + self._download_stock + '.csv'

            if os.path.isfile(path) is False or len(os.listdir(self._folder)) == 0:
                return self.__download()
            else:
                check = input('This file already exist, do you want download proceed [y/n]: ')
                if check.lower() == 'y':
                    return self.__download()
                else: raise Exception("download failed")

    def __delete(self):
        files = glob.glob(self._folder + '*')
        with tqdm(total=len(files), leave=False) as bar:
            for f in files:
                os.remove(f)
                bar.update()

    def __delete_stock(self):
        self._delete_stock = self._delete_stock.upper()
        path = self._folder + self._delete_stock + ".csv"

        ## If file exists, delete it ##
        if os.path.isfile(path):
            os.remove(path)
        else:    ## Show an error ##
            print("Error: %s file not found" % path)
        
    def update(self):
        stock_market = []
        files = glob.glob(self._folder + '*')
        for f in files:
            if f.find('.') != -1:
                f = f.split(self._folder)[-1].split(".")[-2]
            f = f.replace('-', '.')
            stock_market.append(os.path.basename(f))

        with tqdm(total=len(files), leave=False) as bar:
            for file in stock_market:
                bar.update()
                self._update_stock = file
                self.__update_stock()     

    def __update_stock(self):

        self._update_stock = self._update_stock.upper()

        if self._update_stock.find('.') != -1:
            self._update_stock = self._update_stock.replace('.', '-')

        df_1 = pd.read_csv(self._folder + self._update_stock + '.csv', 
                            parse_dates=True, index_col='Date')

        if self._end == "today":
            self._end = date.today().strftime("%Y-%m-%d")
            
        if self._end is None and self._start is not None and df_1.size != 0:
            self._end = df_1.index[-1].strftime("%Y-%m-%d")
                            
        if df_1.size != 0:
            if self._start is None:
                self._start = df_1.index[0]
            else:
                self._start = self._start

            self._end = self._end
            
            yf.pdr_override()

            df = pdr.get_data_yahoo(self._update_stock, 
                                            start=self._start, 
                                            end=self._end,
                                            progress=False,
                                            threads=True,
                                            ignore_tz = True,
                                            )

            df.to_csv(self._folder + self._update_stock + '.csv', 
                            index='Date')

        if self._update is False:
            print(self._update_stock + '.csv', "Downloaded")

    def __main(self):
        path = pathlib.Path(self._folder)
        # Getting the list of directories
        if self._delete:
            self.__delete()
            print("deleted all files from ", path)
            exit()
            
        if path.exists():
            if self._download is True:
                self.__download()
        else:
            if self._download is True:
                path = self._folder 
                os.mkdir(self._folder)
                self.__download()
        
        if self._download_stock is not None:
            self.download_stock()

        if self._update is True:
            self.update()

        if self._update_stock is not None:
            self.__update_stock()

        if self._delete_stock is not None:
            self.__delete_stock()

        dir = os.listdir(path)
        if len(dir) == 0:
            var = input("The directory is empty, you want to download all stock market? [Y/n]")
            if var.lower() == 'y':
                self.__download()
            else:
                raise Exception("If you want to download one stock use --stock action")

class normalize():
    def __init__(self,
                dataframe
                ):

        self.scaler = StandardScaler()
        self.dataframe = dataframe

    def fit_transform(self):
        df_scaled = self.scaler.fit_transform(self.dataframe)
        df_scaled = pd.DataFrame(df_scaled, 
                                columns=self.dataframe.columns,
                                index=self.dataframe.index
                                )
        return df_scaled

    def inverse_transform(self):
        df_inverse = self.scaler.inverse_transform(self.dataframe)
        df_inverse = pd.DataFrame(df_inverse, 
                                columns=self.dataframe.columns,
                                index=self.dataframe.index
                                )
        return df_inverse

    def mean(self):
        df = pd.DataFrame(self.scaler.mean_.reshape(1,-1),
                            columns=self.dataframe.columns,
                            )
        return df['Close'][0]
    
    def std(self):
        df = pd.DataFrame(self.scaler.scale_.reshape(1,-1),
                            columns=self.dataframe.columns,
                            )
        return df['Close'][0]

class StockDataset(torch.utils.data.Dataset):
    def __init__(self, 
                dataframe, 
                sequence_length=0,
                window_size = 7
                ):

        self.dataframe = dataframe

        self.sequence_length = sequence_length
        self.target= "Close"
        self.features = ['High', 'Low', 'Open', 'Volume']

        y = (dataframe[self.target].values)
        x = (dataframe[self.features].values)
        
        self.y = torch.tensor(y).reshape(-1,1).float()
        self.X = torch.tensor(x).float()


    def __len__(self):
        return self.X.shape[0]

    def __getitem__(self, i): 
        if i >= self.sequence_length - 1:
            i_start = i - self.sequence_length + 1
            x = self.X[i_start:(i + 1), :]
        else:
            padding = self.X[0].repeat(self.sequence_length - i - 1, 1)
            x = self.X[0:(i + 1), :]
            x = torch.cat((padding, x), 0)

        if self.sequence_length == 0:
            return self.X[i], self.y[i]

        else: return x, self.y[i] 

def _initTrainDl(x_train: Union[np.ndarray, pd.core.frame.DataFrame]) -> torch.utils.data.DataLoader:
    """
    Initializes the training data loader.

    This method creates a DataLoader object for the training dataset, 
    which is used during the training process. The DataLoader object handles the batching, 
    shuffling, and loading of the training data. The method also sets the batch size, 
    number of workers, and sequence length attributes.

    :param x_train: The training dataset, as a NumPy ndarray or pandas DataFrame
    :type x_train: Union[np.ndarray, pd.core.frame.DataFrame]
    :param batch_size: The batch size to use for training
    :type batch_size: int
    :param num_workers: The number of workers to use for data loading
    :type num_workers: int
    :param sequence_length: The length of the input sequence
    :type sequence_length: int
    :return: The DataLoader object used to handle the training data
    :rtype: torch.utils.data.DataLoader
    """

    train_dl = StockDataset(x_train, sequence_length=training.sequence_length)

    train_dl = DataLoader(train_dl, 
                        batch_size=training.batch_size * (torch.cuda.device_count() \
                                                                   if training.use_cuda else 1),  
                        num_workers=training.num_workers,
                        pin_memory=training.use_cuda,
                        shuffle=True
                        )

    return train_dl

def _initValDl(x_test: Union[np.ndarray, pd.core.frame.DataFrame])-> torch.utils.data.DataLoader:
    """
    Initializes the validation data loader.

    This method creates a DataLoader object for the validation dataset, 
    which is used during the validation process. The DataLoader object handles 
    the batching and loading of the validation data. It uses the batch size, 
    number of workers, and sequence length attributes set during the training process.

    :param x_test: The validation dataset, as a NumPy ndarray or pandas DataFrame
    :type x_test: Union[np.ndarray, pd.core.frame.DataFrame]
    :return: The DataLoader object used to handle the validation data
    :rtype: torch.utils.data.DataLoader
    """

    val_dl = StockDataset(x_test, 
                            sequence_length=training.sequence_length
                            )

    val_dl = DataLoader(val_dl, 
                        batch_size=training.batch_size * (torch.cuda.device_count() \
                                                    if training.use_cuda else 1), 
                        num_workers=training.num_workers,
                        pin_memory=training.use_cuda,
                        shuffle=False
                        )
        
    return val_dl
    
def _initTrainValDl(x_train: Union[np.ndarray, pd.core.frame.DataFrame]) \
                    -> Tuple[torch.utils.data.DataLoader, torch.utils.data.DataLoader]:
    """
    Initializes the training and validation data loaders.

    This method takes the provided training dataset and splits it into training and validation sets. 
    It then initializes DataLoader objects for both sets, which are used during the training and validation 
    processes. The DataLoader objects handle the batching and loading of the data.

    :param x_train: The input dataset to be split into training and validation sets, as a NumPy ndarray or pandas DataFrame
    :type x_train: Union[np.ndarray, pd.core.frame.DataFrame]
    :param validation_sequence: The number of time steps to reserve for validation during training
    :type validation_sequence: int
    :param batch_size: The batch size to use during training
    :type batch_size: int
    :param num_workers: The number of workers to use for data loading
    :type num_workers: int
    :param sequence_length: The length of the input sequence
    :type sequence_length: int
    :return: A tuple containing the DataLoader objects for the training and validation sets
    :rtype: Tuple[torch.utils.data.DataLoader, torch.utils.data.DataLoader]
    """

    scaler = normalize(x_train)

    x_train = scaler.fit_transform()
    val_dl = x_train[-training.validation_sequence:]
    x_train = x_train[:len(x_train)-len(val_dl)]

    train_dl = _initTrainDl(x_train)

    val_dl = _initValDl(val_dl)

    return train_dl, val_dl


import torch
import torch.nn as nn
from torch.optim import SGD, Adam
from torch.utils.data import DataLoader
import torch.optim.lr_scheduler as lr_scheduler
from typing import Union, Tuple

import pyro
from pyro.infer import (
    SVI,
    Trace_ELBO
)
from pyro.optim import ClippedAdam
from tqdm.auto import tqdm
from ..config import shared, training
from ._model import Model

class Trainer(Model):
    def __init__(self, 
                 model = None,
                 **kwargs
                 ):
        
        super().__init__(model, **kwargs)
        self._optimizer = self._initOptimizer()
        pyro.clear_param_store()
        self._optimizer = self._initOptimizer()
        if self.type == "probabilistic":
            self._svi = self._initSVI()
        self._scheduler = self._initScheduler()

    def _initOptimizer(self) -> torch.optim.Optimizer:
        """
        Initializes the optimizer used to train the model.

        This method initializes the optimizer based on the type of the model (probabilistic or neural network). 
        For probabilistic models, a ClippedAdam optimizer is used with the specified learning rate, betas, learning rate decay, and weight decay.
        For neural network models, the Adam optimizer is used with the specified learning rate, betas, epsilon, and weight decay.

        Returns:
            torch.optim.Optimizer: The optimizer instance used to train the model.

        Raises:
            ValueError: If the model type is not recognized.
        """
        if self._model.model_type == "probabilistic":
            adam_params = {"lr": shared.lr, 
                            "betas": shared.betas,
                            "lrd": shared.lrd,
                            "weight_decay": shared.weight_decay
                        }
            return ClippedAdam(adam_params)
        elif self._model.model_type == "neural_network":
            return torch.optim.Adam(self._model.parameters(), 
                                    lr=shared.lr, 
                                    betas=shared.betas, 
                                    eps=shared.eps, 
                                    weight_decay=shared.weight_decay, 
                                    amsgrad=False)
        else:
            raise ValueError("Model type not recognized")
        
    def _initSVI(self) -> pyro.infer.svi.SVI:
        """
        Initializes a Stochastic Variational Inference (SVI) instance to optimize the model and guide.

        This method initializes an SVI instance using the model and guide, the optimizer, and the Trace_ELBO loss function. 
        The SVI instance is used to perform stochastic variational inference, which is an optimization-based approach to approximate 
        posterior distributions for Bayesian models.

        If the model is a BayesianNN, the SVI is initialized with the model directly. For other model types, the SVI is initialized 
        with the model's underlying model attribute.

        Returns:
            pyro.infer.svi.SVI: The SVI instance used to optimize the model and guide.
        """
        if self.name == 'BayesianNN':
            return SVI(self._model, 
                    self._guide, 
                    self._initOptimizer(), 
                    loss=Trace_ELBO()
                    )
        else: 
            return SVI(self._model.model, 
                    self._guide, 
                    self._initOptimizer(), 
                    loss=Trace_ELBO()
                    )
        
    def _initScheduler(self) -> Union[pyro.optim.ExponentialLR, torch.optim.lr_scheduler.StepLR]:
        """
        Initializes a learning rate scheduler to control the learning rate during training.

        This method initializes a learning rate scheduler depending on the model type. For probabilistic models, an instance of
        pyro.optim.ExponentialLR is created, which reduces the learning rate by a factor of gamma for each epoch. For neural network
        models, an instance of torch.optim.lr_scheduler.StepLR is created, which reduces the learning rate by a factor of gamma after 
        every specified step size.

        Raises:
            ValueError: If the model type is not recognized

        Returns:
            Union[pyro.optim.ExponentialLR, torch.optim.lr_scheduler.StepLR]: The learning rate scheduler used to control the learning rate during training.
        """
        if self.type == "probabilistic":
            return pyro.optim.ExponentialLR({'optimizer': self._optimizer, 
                                            'optim_args': {'lr': shared.optim_args}, 
                                            'gamma': shared.gamma}
                                            )
        elif self.type == "neural_network":
            return torch.optim.lr_scheduler.StepLR(self._optimizer, 
                                                step_size=shared.step_size, 
                                                gamma=shared.gamma
                                                )
        else:
            raise ValueError("Model type not recognized")
        
    def _train(self,
               train_dl : torch.utils.data.DataLoader,
               val_dl : torch.utils.data.DataLoader) -> None:

        if self.name!= 'GaussianHMM': 
            self._model.train()

        if self.name == 'DeepMarkovModel':
            self._model.rnn.train()

        best_loss = float('inf')
        counter = 0

        for epoch_ndx in tqdm((range(1, training.epochs + 1)), position=0, leave=True):
            epoch_loss = 0.0
            for x_batch, y_batch in train_dl:   
                if self.type == "neural_network":
                    self._model.train()
                    self._optimizer.zero_grad()  
                    loss = self._computeBatchLoss(x_batch, y_batch)

                    loss.backward()     
                    self._optimizer.step()
                    epoch_loss += loss

                elif self.type == 'probabilistic':
                    loss = self._computeBatchLoss(x_batch, y_batch)
                    epoch_loss += loss

            self._scheduler.step()

            if epoch_ndx % training.validation_cadence != 0:                
                tqdm.write(f"Epoch {epoch_ndx}, Loss: {epoch_loss / len(train_dl)}", 
                           end='\r')

            else:
                val_loss = self._doValidation(val_dl)

                tqdm.write(f"Epoch {epoch_ndx}, Val Loss {val_loss}",
                           end='\r')

                # Early stopping
                stop, best_loss, counter = self._earlyStopping(val_loss, 
                                                               best_loss, 
                                                               counter
                                                               )
                if stop:
                    break  

    def _computeBatchLoss(self, 
                         x_batch : torch.Tensor, 
                         y_batch : torch.Tensor
                         ) -> torch.Tensor:
        """
        Computes the loss for a given batch of data.
        Parameters:
            x_batch (torch.Tensor): the input data
            y_batch (torch.Tensor): the target data
        Returns:
            torch.Tensor: the loss for the given batch of data
        """  
        if self.type == "probabilistic":
            loss = self._svi.step(
                x_data=x_batch,
                y_data=y_batch
            )
        elif self.type == "neural_network":

            x_batch = x_batch.to(self._device)
            y_batch = y_batch.to(self._device)

            output = self._model(x_batch)
            loss_function = nn.MSELoss()
            loss = loss_function(output, y_batch)
        else:
            raise ValueError("Model type not recognized")
        
        return loss
    
    def _doValidation(self, 
                      val_dl : torch.utils.data.DataLoader
                      ) -> float: 
        """
        Performs validation on a given validation data loader.
        Parameters:
            val_dl (torch.utils.data.DataLoader): the validation data loader
        Returns:
            float: the total loss over the validation set
        """

        val_loss = 0.0
        if self.name != 'GaussianHMM': 
            self._model.eval()
        with torch.no_grad():  
            for x_batch, y_batch in val_dl:
                if self.type == 'neural_network':

                    x_batch = x_batch.to(self._device)
                    y_batch = y_batch.to(self._device)

                    output = self._model(x_batch)
                    loss_fn = nn.MSELoss()
                    loss = loss_fn(output, y_batch)
                    val_loss += loss.item()
                elif self.type == 'probabilistic':
                    loss = self._svi.evaluate_loss(x_batch, y_batch)
                    val_loss += loss               

        return val_loss / len(val_dl)

    def _earlyStopping(self,
                       total_loss: float,
                       best_loss: float,
                       counter: int
                       ) -> Tuple[bool, float, int]:
        """
        Implements early stopping during training.

        Parameters:
            total_loss (float): the total validation loss
            best_loss (float): the best validation loss seen so far
            counter (int): the number of epochs without improvement in validation loss
            patience (int): how many epochs to wait for improvement in validation loss before stopping early
            epoch_ndx (int): the current epoch number

        Returns:
            tuple: a tuple containing a bool indicating whether to stop early, the best loss seen so far, and the current counter value
        """

        if total_loss < best_loss:
            best_loss = total_loss
            self._saveModel(self.name)
            counter = 0
        else:
            counter += 1

        if counter >= training.patience:
            print(f"No improvement after {training.patience} epochs. Stopping early.")
            return True, best_loss, counter
        else:
            return False, best_loss, counter


import torch
import torch.nn as nn
from torch.autograd import Variable
from ..config import nn_args, shared, training

class GRU(nn.Module):
    """
    A class representing a Gated Recurrent Unit (GRU) model for stock prediction.

    The GRU model is a type of recurrent neural network (RNN) that is capable of learning long-term dependencies
    in time series data, such as stock prices. The model consists of a GRU layer followed by a series of fully
    connected layers and activation functions.

    :param input_size: The number of input features for the GRU model.
    :type input_size: int
    :param hidden_size: The number of hidden units in each GRU layer.
    :type hidden_size: int
    :param num_layers: The number of GRU layers in the model.
    :type num_layers: int
    :param output_size: The number of output units for the GRU model, corresponding to the predicted target variable(s).
    :type output_size: int
    :param dropout: The dropout percentage applied between layers for regularization, preventing overfitting.
    :type dropout: float
    :example:
        >>> from stockpy.neural_network import GRU
        >>> gru = GRU()
    """
    def __init__(self):
        super().__init__()
        self.gru = nn.GRU(input_size=nn_args.input_size, 
                          hidden_size=nn_args.hidden_size * 2, 
                          num_layers=nn_args.num_layers, 
                          batch_first=True,
                          )
        
        self.layers = nn.Sequential(
            nn.ReLU(),
            nn.Linear(nn_args.hidden_size * 2, nn_args.hidden_size), # [2 * hidden_size] -> [hidden_size]
            nn.ReLU(),
            nn.Dropout(shared.dropout),
            nn.Linear(nn_args.hidden_size, nn_args.input_size), # [hidden_size] -> [input_size]
            nn.ReLU(),
            nn.Dropout(shared.dropout),
            nn.Linear(nn_args.input_size, nn_args.output_size), # [input_size] -> [output_size]
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Defines the forward pass of the GRU model.

        :param x: The input tensor.
        :type x: torch.Tensor

        :returns: The output tensor, corresponding to the predicted target variable(s).
        :rtype: torch.Tensor
        """
        batch_size = x.size(0)
        h0 = Variable(torch.zeros(nn_args.num_layers, 
                                  batch_size, 
                                  nn_args.hidden_size * 2)).to(training.device)
        
        _, (hn) = self.gru(x, (h0))
        out = self.layers(hn[0])       
        out = out.view(-1, 1)

        return out
    
    def to(self, device: torch.device) -> None:
        """
        Moves the model to the specified device.

        :param device: The device to move the model to.
        :type device: torch.device
        """
        super().to(device)

    @property
    def model_type(self) -> str:
        """
        Returns the type of model.

        :returns: The model type as a string.
        :rtype: str
        """
        return "neural_network"


import torch
import torch.nn as nn
from ..config import nn_args, shared

class MLP(nn.Module):
    """
    A class representing a Multilayer Perceptron (MLP) neural network model for stock prediction.

    The MLP model consists of a series of linear layers, each followed by a non-linear activation function,
    such as ReLU, and dropout for regularization. The model is designed to learn non-linear relationships
    between input features and the target variable for predicting stock prices.

    :param input_size: The number of input features for the MLP model.
    :type input_size: int
    :param hidden_size: The number of hidden units in each hidden layer of the MLP model.
    :type hidden_size: int
    :param num_layers: The number of layers in the MLP model, including input, hidden, and output layers.
    :type num_layers: int
    :param output_dim: The number of output units for the MLP model, corresponding to the predicted target variable(s).
    :type output_dim: int
    :param dropout: The dropout percentage applied between layers for regularization, preventing overfitting.
    :type dropout: float
    :example:
        >>> from stockpy.neural_network import MLP
        >>> mlp = MLP()
    """
    def __init__(self):
        """
        Initializes the MLP neural network model.

        :param args: The arguments to configure the model.
        :type args: ModelArgs
        """
        super().__init__()
        self.layers = nn.Sequential(
            nn.Linear(nn_args.input_size, nn_args.hidden_size),   # [input_size] -> [hidden_size]
            nn.ReLU(),
            nn.Dropout(shared.dropout),
            nn.Linear(nn_args.hidden_size, nn_args.hidden_size),  # [hidden_size] -> [hidden_size]
            nn.ReLU(),
            nn.Dropout(shared.dropout),
            nn.Linear(nn_args.hidden_size, nn_args.input_size),   # [hidden_size] -> [input_size]
            nn.ReLU(),
            nn.Dropout(shared.dropout),
            nn.Linear(nn_args.input_size, nn_args.output_size),   # [input_size] -> [output_size]
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Defines the forward pass of the neural network.

        :param x: The input tensor.
        :type x: torch.Tensor

        :returns: The output tensor, corresponding to the predicted target variable(s).
        :rtype: torch.Tensor
        """
        return self.layers(x)

    @property
    def model_type(self) -> str:
        """
        Returns the type of model.

        :returns: The model type as a string.
        :rtype: str
        """
        return "neural_network"